from athina.interfaces.model import Model
from ..ragas_evaluator import RagasEvaluator
from athina.evals.eval_type import RagasEvalTypeId
from athina.metrics.metric_type import MetricType
from ragas.metrics import context_precision
from typing import List, Optional

"""
RAGAS Context Precision Docs: https://docs.ragas.io/en/latest/concepts/metrics/context_precision.html
RAGAS Context Precision Github: https://github.com/explodinggradients/ragas/blob/main/src/ragas/metrics/_context_precision.py
"""


class RagasContextPrecision(RagasEvaluator):
    """
    This evaluator calculates the precision of the context with respect to the expected response.
    Context Precision is a metric that evaluates whether all of the ground-truth relevant items present in the context are ranked higher or not.
    Ideally all the relevant chunks must appear at the top ranks.
    """

    @property
    def name(self):
        return RagasEvalTypeId.RAGAS_CONTEXT_PRECISION.value

    @property
    def display_name(self):
        return "Ragas Context Precision"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.RAGAS_CONTEXT_PRECISION.value]

    @property
    def ragas_metric(self):
        return context_precision

    @property
    def ragas_metric_name(self):
        return "context_precision"

    @property
    def default_model(self):
        return Model.GPT35_TURBO.value

    @property
    def required_args(self):
        return ["query", "context", "expected_response"]

    @property
    def examples(self):
        return None

    @property
    def grade_reason(self) -> str:
        return "This metric evaluates whether all of the ground-truth relevant items present in the context are ranked higher or not. Ideally all the relevant chunks must appear at the top ranks"

    def is_failure(self, score) -> Optional[bool]:
        return (
            bool(score < self._failure_threshold)
            if self._failure_threshold is not None
            else None
        )

    def generate_data_to_evaluate(
        self, context, query, expected_response, **kwargs
    ) -> dict:
        """
        Generates data for evaluation.

        :param context: list of strings of retrieved context
        :param query: user query
        :param expected_response: expected output
        :return: A dictionary with formatted data for evaluation
        """
        data = {
            "contexts": [context],
            "question": [query],
            "ground_truths": [[expected_response]],
        }
        return data



SYSTEM_MESSAGE = """
You are an expert at determining whether a user's question was addressed / resolved by the AI or not. 
If the user is asking a question, it is considered resolved if the AI provides a clear answer to the question.
If the user is making a statement, it is considered resolved if the AI provides a clear response to the statement.
"""

USER_MESSAGE = """
- Consider the provided conversation messages.
- For each user message, determine whether the AI's response addressed the user's message or not.
- If the AI's response addressed the user's message, mark it as "Resolved".
- If the AI's response did not address the user's message, mark it as "Unresolved".
- If the AI's response partially addressed the user's message, mark it as "Partial".

Return a JSON array of objects with the following structure:
{{
    "details": [{{
        "message": "<User message>",
        "resolution": "Resolved/Unresolved/Partial"
        "explanation": "Explain why the AI's response addressed the user's message or not."
    }}]
}}

Here are the conversation messages to consider:
{messages}
"""



# Step to make a call to pinecone index to fetch relevent chunks
from typing import Optional, Union, Dict, Any

from pydantic import PrivateAttr
from athina.steps import Step
from jinja2 import Environment
from llama_index.vector_stores import QdrantVectorStore
from llama_index import VectorStoreIndex
from llama_index.retrievers import VectorIndexRetriever
import qdrant_client


class QdrantRetrieval(Step):
    """
    Step that makes a call to qdrant index to fetch relevant chunks.

    Attributes:
        collection_name: collection name in qdrant
        url: url of the qdrant server
        top_k: How many chunks to fetch.
        api_key: api key for the qdrant server
        input_column: column name in the input data
        env: jinja environment
    """

    collection_name: str
    url: str
    top_k: int
    api_key: str
    input_column: str
    env: Environment = None
    _qdrant_client: qdrant_client.QdrantClient = PrivateAttr()
    _vector_store: QdrantVectorStore = PrivateAttr()
    _vector_index: VectorStoreIndex = PrivateAttr()
    _retriever: VectorIndexRetriever = PrivateAttr()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._qdrant_client = qdrant_client.QdrantClient(
            url=self.url, api_key=self.api_key
        )
        self._vector_store = QdrantVectorStore(
            client=self._qdrant_client, collection_name=self.collection_name
        )
        self._vector_index = VectorStoreIndex.from_vector_store(
            vector_store=self._vector_store
        )
        self._retriever = VectorIndexRetriever(
            index=self._vector_index, similarity_top_k=self.top_k
        )

    class Config:
        arbitrary_types_allowed = True

    def execute(self, input_data: Any) -> Union[Dict[str, Any], None]:
        """makes a call to pinecone index to fetch relevent chunks"""

        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict):
            raise TypeError("Input data must be a dictionary.")

        input_text = input_data.get(self.input_column, None)

        if input_text is None:
            return None

        try:
            response = self._retriever.retrieve(input_text)
            result = [node.get_content() for node in response]
            return {
                "status": "success",
                "data": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "data": str(e),
            }



import os
import re
import json
import requests
from typing import Any, Dict, Optional, Tuple, Union
from athina.evals.grounded.similarity import CosineSimilarity
from athina.errors.exceptions import NoOpenAiApiKeyException
from athina.helpers.jinja_helper import PreserveUndefined
from athina.helpers.json import extract_json_path, validate_json
from athina.helpers.logger import logger
from athina.keys.openai_api_key import OpenAiApiKey
from athina.llms.openai_service import OpenAiService
from athina.steps.code_execution import CodeExecution
import subprocess
import tempfile
from jinja2 import Environment


def _standardize_url(url):
    """
    Generate a standardized URL by adding 'http://' if it's missing.

    Args:
        url (str): The input URL to be standardized.

    Returns:
        str: The standardized URL.
    """
    if url.startswith("http://") or url.startswith("https://"):
        return url
    else:
        return "http://" + url


def _preprocess_strings(keywords, text, case_sensitive):
    """
    Preprocess the keywords based on the case_sensitive flag.

    Args:
        keywords (str or List[str]): The keyword(s) to preprocess.
        case_sensitive (bool): Whether the preprocessing should be case-sensitive.

    Returns:
        List[str]: The preprocessed keywords.
    """
    # If keywords is a string, convert it to a list
    if isinstance(keywords, str):
        keywords = keywords.split(",")

    # Strip leading and spaces from the keywords
    keywords = list(map(lambda k: k.strip(), keywords))

    # If case_sensitive is False, convert all keywords and text to lowercase
    if not case_sensitive:
        keywords = [keyword.lower() for keyword in keywords]
        text = text.lower()

    return keywords, text


def regex(pattern, text, **kwargs):
    """
    Perform a regex search on the text and return a dictionary indicating whether the pattern was found.

    Args:
        pattern (str): The regex pattern to search for.
        text (str): The text string to search within.

    Returns:
        dict: A dictionary containing the result of the regex search and the reason for the result.
    """
    match = re.search(pattern, text)
    if match:
        return {"result": True, "reason": f"regex pattern {pattern} found in output"}
    else:
        return {
            "result": False,
            "reason": f"regex pattern {pattern} not found in output",
        }


def contains_any(keywords, text: str, case_sensitive=False, **kwargs):
    """
    Check if any of the provided keywords are present in the text.

    Args:
        keywords (str or List[str]): The keyword(s) to search for in the text.
        text (str): The text string to search within.
        case_sensitive (bool, optional): Whether the search should be case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the search and the reason for the result.
    """
    keywords, text = _preprocess_strings(keywords, text, case_sensitive)
    found_keywords = []
    for keyword in keywords:
        if keyword in text:
            found_keywords.append(keyword)

    if found_keywords:
        result = True
        reason = f"One or more keywords were found in output: " + ", ".join(
            found_keywords
        )
    else:
        result = False
        reason = "No keywords found in output"

    return {"result": result, "reason": reason}


def contains_all(keywords, text, case_sensitive=False, **kwargs):
    """
    Check if all the provided keywords are present in the text.

    Args:
        keywords (List[str]): The list of keywords to search for in the text.
        text (str): The text string to search within.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the keyword search and the reason for the result.
    """
    keywords, text = _preprocess_strings(keywords, text, case_sensitive)
    missing_keywords = []
    for keyword in keywords:
        if keyword not in text:
            result = False
            missing_keywords.append(keyword)
    if (len(missing_keywords)) > 0:
        result = False
        reason = f"keywords not found in output: " + ", ".join(missing_keywords)
    else:
        result = True
        reason = f"{len(keywords)}/{len(keywords)} keywords found in output"

    return {"result": result, "reason": reason}


def contains(keyword, text, case_sensitive=False, **kwargs):
    """
    Check if the text contains a specific keyword.

    Args:
        keyword (str): The keyword to search for in the text.
        text (str): The text string to search within.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the keyword search and the reason for the result.
    """
    if case_sensitive == False:
        text = text.lower()
        keyword = keyword.lower()
    if keyword not in text:
        result = False
        reason = f"keyword not found in output: " + keyword
    else:
        result = True
        reason = f"keyword {keyword} found in output"

    return {"result": result, "reason": reason}


def contains_none(keywords, text, case_sensitive=False, **kwargs):
    """
    Check if none of the provided keywords are present in the text.

    Args:
        keywords (str or List[str]): The keyword(s) to search for in the text.
        text (str): The text string to search within.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the check and the reason for the result.
    """
    keywords, text = _preprocess_strings(keywords, text, case_sensitive)
    found_keywords = []
    for keyword in keywords:
        if keyword in text:
            found_keywords.append(keyword)

    if found_keywords:
        result = False
        reason = f"One or more keywords were found in output: " + ", ".join(
            found_keywords
        )
    else:
        result = True
        reason = "No keywords found in output"

    return {"result": result, "reason": reason}


def contains_json(text, **kwargs):
    """
    Check if the text contains valid JSON.

    Args:
        text (str): The text string to check for valid JSON.

    Returns:
        dict: A dictionary containing the result of the JSON check and the reason for the result.
    """
    trimmed_output = text.strip()
    pattern = (
        r'\{(?:\s*"(?:\\.|[^"\\])*"\s*:\s*(?:"(?:\\.|[^"\\])*"|[^{}\[\]:,]+)|[^{}]+)*\}'
    )
    matches = re.findall(pattern, trimmed_output)

    if matches:
        results = []
        errors = []
        for potential_json_string in matches:
            try:
                parsed_json = json.loads(potential_json_string)
                results.append({"json": parsed_json, "valid": True})
            except json.JSONDecodeError as e:
                errors.append(
                    {"json": potential_json_string, "valid": False, "error": str(e)}
                )
        if errors:
            return {
                "result": False,
                "reason": "Output contains a potential JSON but it is invalid",
                "matches": results,
                "errors": errors,
            }
        else:
            return {
                "result": True,
                "reason": "Output contains JSON",
                "matches": results,
            }
    else:
        return {"result": False, "reason": "Output does not contain JSON"}


def contains_email(text, **kwargs):
    """
    Check if the text contains an email address.

    Args:
        text (str): The text string to check for an email address.

    Returns:
        dict: A dictionary containing the result of the email address check and the reason for the result.
    """
    return regex(r"[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+", text)


def is_json(text, **kwargs):
    """
    Check if the text contains valid JSON.

    Args:
        text (str): The text string to check for valid JSON.

    Returns:
        dict: A dictionary containing the result of the JSON check and the reason for the result.
    """
    try:
        json.loads(text)
        result = True
    except json.JSONDecodeError:
        result = False
    if result:
        return {
            "result": True,
            "reason": "Output contains JSON",
        }
    else:
        return {
            "result": False,
            "reason": "Output does not contain JSON",
        }


def is_email(text, **kwargs):
    """
    Check if the text is a valid email address.

    Args:
        text (str): The text string to check for a valid email address.

    Returns:
        dict: A dictionary containing the result of the email address check and the reason for the result.
    """
    return regex(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$", text)


def contains_link(text, **kwargs):
    """
    Check if the text contains a link.

    Args:
        text (str): The text string to check for a link.

    Returns:
        dict: A dictionary containing the result of the link check and the reason for the result.
    """
    pattern = r"(?!.*@)(?:https?://)?(?:www\.)?\S+\.\S+"
    result = bool(re.search(pattern, text))
    if result:
        return {"result": True, "reason": "Link found in output"}
    else:
        return {"result": False, "reason": "No link found in output"}


def contains_valid_link(text, **kwargs):
    """
    Check if the text contains a valid link.

    Args:
        text (str): The text string to check for a valid link.

    Returns:
        dict: A dictionary containing the result of the link check and the reason for the result.
    """
    pattern = r"(?!.*@)(?:https?://)?(?:www\.)?\S+\.\S+"
    link_match = re.search(pattern=pattern, string=text)
    if link_match:
        matched_url = link_match.group()
        if matched_url:
            standardized_url = _standardize_url(matched_url)
            try:
                text = requests.head(standardized_url)
                if text.status_code == 200:
                    return {
                        "result": True,
                        "reason": f"link {matched_url} found in output and is valid",
                    }
                else:
                    return {
                        "result": False,
                        "reason": f"link {matched_url} found in output but is invalid",
                    }
            except:
                return {
                    "result": False,
                    "reason": f"link {matched_url} found in output but is invalid",
                }
    return {"result": False, "reason": f"no link found in output"}


def no_invalid_links(text, **kwargs):
    """
    Check for invalid links in the text.

    Args:
        text (str): The text string to check for invalid links.

    Returns:
        dict: A dictionary containing the result of the link check and the reason for the result.
    """
    pattern = r"(?!.*@)(?:https?://)?(?:www\.)?\S+\.\S+"
    link_match = re.search(pattern=pattern, string=text)
    if link_match:
        matched_url = link_match.group()
        if matched_url:
            standardized_url = _standardize_url(matched_url)
            try:
                text = requests.head(standardized_url)
                if text.status_code == 200:
                    return {
                        "result": True,
                        "reason": f"link {matched_url} found in output and is valid",
                    }
                else:
                    return {
                        "result": False,
                        "reason": f"link {matched_url} found in output but is invalid",
                    }
            except:
                return {
                    "result": False,
                    "reason": f"link {matched_url} found in output but is invalid",
                }
    return {"result": True, "reason": f"no invalid link found in output"}


def api_call(
    url: str,
    response: str,
    query: Optional[str] = None,
    context: Optional[str] = None,
    expected_response: Optional[str] = None,
    payload: dict = None,
    headers: dict = None,
):
    """
    Make an API call with payload to the specified URL.

    Args:
        url (str): The URL to make the API call to.
        text (str): The text to be added to the payload.
        query (Optional[str]): The query parameter to be added to the payload.
        context (Optional[str]): The context parameter to be added to the payload.
        expected_response (Optional[str]): The expected text parameter to be added to the payload.
        payload (dict, optional): The payload to be sent in the API call. Defaults to None.
        headers (dict, optional): The headers to be included in the API call. Defaults to None.

    Returns:
        dict: A dictionary containing the result and reason of the API call.
    """
    if payload is None:
        payload = {}
    if headers is None:
        headers = {}
    payload["response"] = response
    if query:
        payload["query"] = query
    if context:
        payload["context"] = context
    if expected_response:
        payload["expected_response"] = expected_response
    # Check the status code and set the reason accordingly
    try:
        api_response = requests.post(url, json=payload, headers=headers)
        if api_response.status_code == 200:
            # Success
            result = api_response.json().get("result")
            reason = api_response.json().get("reason")
        elif api_response.status_code == 400:
            # Bad Request
            result = False
            reason = "Bad Request: The server could not understand the request due to invalid syntax."
        elif api_response.status_code == 401:
            # Unauthorized
            result = False
            reason = "Unauthorized: Authentication is required and has failed or has not been provided."
        elif api_response.status_code == 500:
            # Internal Server Error
            result = False
            reason = (
                "Internal Server Error: The server encountered an unexpected condition."
            )
        else:
            # Other error codes
            result = False
            reason = f"An error occurred: {api_response.status_code}"
    except Exception as e:
        # Handle any exceptions that occur during the API call
        result = False
        reason = f"API Request Exception: {e}"

    return {"result": result, "reason": reason}


def equals(expected_text, text, case_sensitive=False, **kwargs):
    """
    Check if the text exactly matches the expected text.

    Args:
        expected_text (str): The expected text to compare against.
        text (str): The text to compare with the expected output.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result and reason of the comparison.
    """
    if case_sensitive == False:
        text = text.lower()
        expected_text = expected_text.lower()
    if text == expected_text:
        result = True
        reason = "✅ Text exactly matches expected text"
    else:
        result = False
        reason = "output does not exactly match expected text"
    return {"result": result, "reason": reason}


def starts_with(substring, text, case_sensitive=False, **kwargs):
    """
    Check if the text starts with a specified substring.

    Args:
        substring (str): The substring to check for at the start of the text.
        text (str): The text string to check.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the check and the reason for the result.
    """
    if case_sensitive == False:
        text = text.lower()
        substring = substring.lower()
    result = text.startswith(substring)
    if result == True:
        return {"result": result, "reason": "output starts with " + substring}
    else:
        return {"result": result, "reason": "output does not start with " + substring}


def ends_with(substring, text, case_sensitive=False, **kwargs):
    """
    Check if the text ends with a specified substring.

    Args:
        substring (str): The substring to check for at the end of the text.
        text (str): The text string to check.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the check and the reason for the result.
    """
    if case_sensitive == False:
        text = text.lower()
        substring = substring.lower()
    result = text.endswith(substring)
    if result == True:
        return {"result": result, "reason": "output ends with " + substring}
    else:
        return {"result": result, "reason": "output does not end with " + substring}


def length_less_than(max_length, text, **kwargs):
    """
    Check if the length of the text is less than a specified maximum length.

    Args:
        max_length (int): The maximum length that the text should have.
        text (str): The text string to check the length of.

    Returns:
        dict: A dictionary containing the result of the length check and the reason for the result.
    """
    if len(text) < max_length:
        return {
            "result": True,
            "reason": f"output length is less than {max_length} characters",
        }
    else:
        return {
            "result": False,
            "reason": f"output length is greater than {max_length} characters",
        }


def length_greater_than(min_length, text, **kwargs):
    """
    Check if the length of the text is greater than a specified minimum length.

    Args:
        min_length (int): The minimum length that the text should have.
        text (str): The text string to check the length of.

    Returns:
        dict: A dictionary containing the result of the length check and the reason for the result.
    """
    if len(text) > min_length:
        return {
            "result": True,
            "reason": f"output length is greater than {min_length} characters",
        }
    else:
        return {
            "result": False,
            "reason": f"output length is less than {min_length} characters",
        }


def length_between(min_length, max_length, text, **kwargs):
    """
    Check if the length of the text is between a specified minimum and maximum length.

    Args:
        min_length (int): The minimum length that the text should have.
        max_length (int): The maximum length that the text should have.
        text (str): The text string to check the length of.

    Returns:
        dict: A dictionary containing the result of the length check and the reason for the result.
    """
    if min_length <= len(text) <= max_length:
        return {
            "result": True,
            "reason": f"output length is between {min_length} and {max_length} characters",
        }
    else:
        return {
            "result": False,
            "reason": f"output length is not between {min_length} and {max_length} characters",
        }


def one_line(text, **kwargs):
    """
    Check if the text is a single line.

    Args:
        text (str): The text string to check.

    Returns:
        dict: A dictionary containing the result of the check and the reason for the result.
    """
    if "\n" in text or len(text.splitlines()) > 1:
        return {"result": False, "reason": "output contains multiple lines"}
    else:
        return {"result": True, "reason": "output is a single line"}


def json_schema(actual_json: Union[dict, str], **kwargs) -> Dict[str, Any]:
    """
    Check if the actual_json matched the schema definition.

    Args:
        actual_json (dict or str): The JSON string to check with the schema.
    """
    try:
        # Load the actual JSON data from the input
        actual_json = _load_json(actual_json)

        # Retrieve the schema from the provided keyword arguments
        schema = _get_schema(kwargs)
        if not schema:
            # Return failure if schema is not provided
            return {"result": False, "reason": "Schema not provided"}

        # Validate the actual JSON against the schema
        passed, reason = _validate_json_with_schema(actual_json, schema)
        if not passed:
            # Return failure if validation does not pass
            return {"result": False, "reason": reason}

        # Return success if validation passes
        return {"result": True, "reason": "JSON schema passed"}
    except Exception as e:
        # Log and raise any exceptions that occur during the process
        logger.error(f"Error occurred during JSON schema validation: {e}")
        raise e


def json_validation(
    actual_json: Union[dict, str], expected_json: Union[dict, str], **kwargs
) -> Dict[str, Any]:
    """
    Check if the actual JSON and expected JSON match the validation rules.

    Args:
        actual_json (dict or str): The actual JSON string to compare against the expected JSON.
        expected_json (dict or str): The expected JSON string to compare against the actual JSON.
    """
    try:
        actual_json = _load_json(actual_json)
        expected_json = _load_json(expected_json)

        validations = kwargs.get("validations", [])
        if validations:
            for validation in validations:
                validation_result = _apply_validation(
                    actual_json, expected_json, validation
                )
                validation_passed = validation_result[0]
                validation_reason = validation_result[1]
                if not validation_passed:
                    return {"result": False, "reason": validation_reason}

        return {"result": True, "reason": "Json validation passed"}
    except Exception as e:
        logger.error(f"Error occurred during Json validation eval: {e}")
        raise e


def _bandit_check(code: str) -> None:
    """
    Run Bandit security check on the provided code.
    """
    with tempfile.NamedTemporaryFile(delete=False, suffix=".py") as temp_file:
        temp_file.write(code.encode("utf-8"))
        temp_file_path = temp_file.name
    try:
        result = subprocess.run(
            ["bandit", "-r", temp_file_path, "-f", "json", "-c", "bandit.yml"],
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            return json.dumps(result.stdout)
    finally:
        os.remove(temp_file_path)
    return None


def custom_code_eval(code, **kwargs):
    """
    Run custom code provided by the user.

    Args:
        code (str): The custom code to run.

    Returns:
        dict: A dictionary containing the result of the check and the reason for the result.
    """
    # Create an instance of CodeExecution
    code_execution = CodeExecution(code=code)

    # Execute the code using the CodeExecution instance
    result = code_execution.execute(kwargs)

    # Check the result and return the appropriate response
    if result.get("status") == "success":
        if result.get("data"):
            return {"result": True, "reason": "Custom eval code passed"}
        else:
            return {"result": False, "reason": "Custom eval code failed"}
    else:
        return {
            "result": False,
            "reason": result.get("data", "Error in custom eval code eval"),
        }


def _load_json(json_data: Union[dict, str]) -> dict:
    if isinstance(json_data, str):
        return json.loads(json_data)
    return json_data


def _get_schema(kwargs: Dict[str, Any]) -> dict:
    schema = kwargs.get("schema")
    if schema and isinstance(schema, str):
        return json.loads(schema.replace("\n", "").replace("\t", ""))
    return schema


def _validate_json_with_schema(json_data: dict, schema: dict) -> Tuple[bool, str]:
    return validate_json(json_data, schema)


def _apply_validation(actual_json: dict, expected_json: dict, validation: dict) -> bool:
    validating_function = validation.get("validating_function")
    json_path = validation.get("json_path")
    actual_value = extract_json_path(actual_json, json_path)
    expected_value = extract_json_path(expected_json, json_path)

    if validating_function == "Equals":
        return _validate_equals(actual_value, expected_value, validation, json_path)
    elif validating_function == "Cosine Similarity":
        return _validate_cosine_similarity(
            actual_value, expected_value, validation, json_path
        )
    elif validating_function == "LLM Similarity":
        return _validate_llm_similarity(
            actual_value, expected_value, validation, json_path
        )
    else:
        error_message = f"Validation function {validating_function} not supported"
        logger.error(error_message)
        return False, error_message


def _validate_equals(
    actual_value: Any, expected_value: Any, validation: dict, json_path: str
) -> bool:
    case_sensitive = validation.get("case_sensitive", False)
    if (
        not case_sensitive
        and isinstance(actual_value, str)
        and isinstance(expected_value, str)
    ):
        actual_value = str(actual_value).lower()
        expected_value = str(expected_value).lower()
    if actual_value != expected_value:
        error_message = f"JSON path {json_path} does not match expected value"
        logger.error(error_message)
        return False, error_message
    return True, None


def _validate_cosine_similarity(
    actual_value: str, expected_value: str, validation: dict, json_path: str
) -> bool:
    threshold = validation.get("pass_threshold", 0.8)
    cosine_similarity = CosineSimilarity().compare(
        str(actual_value), str(expected_value)
    )
    if cosine_similarity < threshold:
        error_message = f"Cosine similarity score of {round(cosine_similarity, 2)} for {json_path} is less than the threshold ({threshold})."
        logger.error(error_message)
        return False, error_message
    return True, None


def _validate_llm_similarity(
    actual_value: str, expected_value: str, validation: dict, json_path: str
) -> bool:
    open_ai_api_key = (
        validation.get("open_ai_api_key")
        or OpenAiApiKey.get_key()
        or os.environ.get("OPENAI_API_KEY")
    )
    if not open_ai_api_key:
        raise NoOpenAiApiKeyException()

    OpenAiApiKey.set_key(open_ai_api_key)
    llm_service = OpenAiService()
    messages = _get_messages(validation, actual_value, expected_value)

    response = llm_service.json_completion(
        model=validation.get("model", "gpt-3.5-turbo"),
        messages=messages,
        temperature=0.0,
    )

    try:
        result = response["result"]
        explanation = response["explanation"]
        if bool(str(result).lower() == "fail"):
            error_message = f"LLM Similarity validation failed for {json_path}. Reason: {explanation}"
            logger.error(error_message)
            return False, error_message
        return True, None
    except Exception as e:
        error_message = (
            f"Error occurred during LLM similarity validation for {json_path}"
        )
        logger.error(error_message)
        return False, error_message


def _get_messages(validation: dict, actual_value: Any, expected_value: Any) -> list:
    if validation.get("system_message") and validation.get("user_message"):
        env = Environment(
            variable_start_string="{{",
            variable_end_string="}}",
            undefined=PreserveUndefined,
        )
        render_context = {"actual": actual_value, "expected": expected_value}
        system_message = env.from_string(validation.get("system_message")).render(
            render_context
        )
        user_message = env.from_string(validation.get("user_message")).render(
            render_context
        )
        return [
            {"role": "system", "content": system_message},
            {"role": "user", "content": user_message},
        ]
    else:
        # Default messages
        system_message = """
        You are an expert at evaluating whether two given strings are similar or not. Consider semantic similarity also while evaluating.
        You MUST return a JSON object with the following fields: 
        - result: Result must be either 'Pass' or 'Fail'.
        - explanation: An explanation of why the result is Pass or Fail.
        - score: Any matching score you have used to come to the result.
        """

        user_message = f"""
        Following are two strings:
        1. String 1: {actual_value}.
        2. String 2: {expected_value}.
        """

        return [
            {"role": "system", "content": system_message},
            {"role": "user", "content": user_message},
        ]


"""
A dictionary containing the available operations and their corresponding functions.
"""
operations = {
    "Regex": regex,
    "ContainsAny": contains_any,
    "ContainsAll": contains_all,
    "Contains": contains,
    "ContainsNone": contains_none,
    "ContainsJson": contains_json,
    "ContainsEmail": contains_email,
    "IsJson": is_json,
    "IsEmail": is_email,
    "NoInvalidLinks": no_invalid_links,
    "ContainsLink": contains_link,
    "ContainsValidLink": contains_valid_link,
    "Equals": equals,
    "StartsWith": starts_with,
    "EndsWith": ends_with,
    "LengthLessThan": length_less_than,
    "LengthGreaterThan": length_greater_than,
    "LengthBetween": length_between,
    "ApiCall": api_call,
    "OneLine": one_line,
    "JsonSchema": json_schema,
    "JsonValidation": json_validation,
    "CustomCodeEval": custom_code_eval,
}



from typing import Tuple, List, Optional
from athina.llms.abstract_llm_service import AbstractLlmService
from .question_answerer import QuestionAnswerer, QuestionAnswererResponse
from athina.llms.openai_service import OpenAiService


class QuestionAnswererChainOfThought(QuestionAnswerer):

    _llm_service: AbstractLlmService

    """
    This class responds to a list of closed-ended (Y/N) questions based on a provided context.
    It does so using a separate LLM inference call with CoT prompting for each question.
    It also asks the LLM to provide an explanation for each answer, which helps improve the reasoning.
    """

    # Pre-defined prompts for OpenAI's GPT model
    SYSTEM_MESSAGE = """ 
        You are an expert at responding to closed-ended (Yes/No) questions using the provided context.
    """

    USER_MESSAGE_TEMPLATE = """
        Let's think step by step.
        1. Consider the following: 
           Question: {}.
           Context: {}.
        2. Based on the context provided, think through the question and determine an explanation for your response.
        3. If you cannot determine an answer, respond with 'Unknown'.
        4. Respond to the question with an explanation, leading up to a final answer to the question: 'Yes', 'No', or 'Unknown'.
        5. Return a JSON object in the following format: "answer": "answer", "explanation": "explanation"
    """

    def __init__(
        self,
        model: str = "gpt-4-1106-preview",
        llm_service: Optional[AbstractLlmService] = None,
    ):
        """
        Initialize the QuestionAnswerer class.
        """
        self._model = model

        if llm_service is None:
            self._llm_service = OpenAiService()
        else:
            self._llm_service = llm_service

    def answer(self, questions: List[str], context: str) -> Tuple[dict, dict]:
        """
        Respond to each question from the provided 'questions' given the context.
        """

        results = {}
        simple_output = {}
        for question in questions:
            try:
                response = self.answer_question(question, context)
                results[question] = response
                simple_output[question] = response["answer"]
            except:
                results[question] = {
                    "answer": "Error",
                    "explanation": None,
                }
                simple_output[question] = "Error"
        return results, simple_output

    def answer_question(self, question: str, context: str) -> QuestionAnswererResponse:
        """
        Respond to each question from the provided 'questions' given the context.

        Args:
            question (str): A set of questions posed to the chatbot.
            context (str): Context used to inform the chatbot's answers.

        Returns:
            dict: Evaluation results formatted as a dictionary with questions as keys and
                  'Yes', 'No', or 'Unknown' as values.
        """

        user_message = self.USER_MESSAGE_TEMPLATE.format(question, context)
        messages = [
            {"role": "system", "content": self.SYSTEM_MESSAGE},
            {"role": "user", "content": user_message},
        ]

        # Extract JSON object from LLM response
        json_response = self._llm_service.json_completion(
            model=self._model,
            messages=messages,
        )

        if json_response is None:
            raise Exception("No response from LLM")

        answer = json_response["answer"]
        explanation = json_response["explanation"]

        return {
            "answer": answer,
            "explanation": explanation,
        }



from typing import List, Optional
from athina.interfaces.athina import AthinaFilters
from athina.interfaces.data import DataPoint
from .base_loader import BaseLoader
from athina.services.athina_api_service import AthinaApiService
from dataclasses import asdict


class ResponseLoader(BaseLoader):
    """
    This class is a data loader for evals that only evaluate the response.

    Attributes:
        col_response (str): The column name corresponding to the response.
        raw_dataset (dict): The raw dataset as loaded from the source.
        processed_dataset (list): The processed dataset with responses.
    """

    def __init__(
        self,
        col_response: str = "response",
        col_query: Optional[str] = "query",
        col_context: Optional[str] = "context",
        col_expected_response: Optional[str] = "expected_response",
    ):
        """
        Initializes the loader with specified or default column names.
        """
        self.col_response = col_response
        self.col_query = col_query
        self.col_context = col_context
        self.col_expected_response = col_expected_response
        self._raw_dataset = {}
        self._processed_dataset: List[DataPoint] = []

    def process(self) -> None:
        """
        Transforms the raw data into a structured format. Processes each entry from the raw dataset, and extracts attributes.

        Raises:
            KeyError: If mandatory columns (response) are missing in the raw dataset.
        """
        for raw_instance in self._raw_dataset:
            # Check for mandatory columns in raw_instance
            if self.col_response not in raw_instance:
                raise KeyError(f"'{self.col_response}' not found in provided data.")
            if self.col_query in raw_instance and not isinstance(
                raw_instance.get(self.col_query), str
            ):
                raise TypeError(f"'{self.col_query}' is not of type string.")
            if self.col_context in raw_instance and not isinstance(
                raw_instance.get(self.col_context), str
            ):
                raise TypeError(f"'{self.col_context}' is not of type string.")
            if self.col_expected_response in raw_instance and not isinstance(
                raw_instance.get(self.col_expected_response), str
            ):
                raise TypeError(
                    f"'{self.col_expected_response}' is not of type string."
                )
            # Create a processed instance with mandatory fields
            processed_instance = {
                "response": raw_instance[self.col_response],
                "query": raw_instance.get(self.col_query, None),
                "context": raw_instance.get(self.col_context, None),
                "expected_response": raw_instance.get(self.col_expected_response, None),
            }
            # removing keys with None values
            processed_instance = {
                k: v for k, v in processed_instance.items() if v is not None
            }
            # Store the results
            self._processed_dataset.append(processed_instance)

    def load_athina_inferences(
        self,
        filters: Optional[AthinaFilters] = None,
        limit: Optional[int] = None,
    ):
        """
        Load data from Athina API.
        """
        self._raw_dataset = AthinaApiService.fetch_inferences(
            filters=filters, limit=limit
        )
        for raw_dataset in self._raw_dataset:
            raw_dataset_dict = asdict(raw_dataset)
            processed_instance = {
                "response": raw_dataset_dict["prompt_response"],
            }
            self._processed_dataset.append(processed_instance)
        return self._processed_dataset



from typing import List, Optional

from athina.interfaces.model import Model
from ..ragas_evaluator import RagasEvaluator
from athina.evals.eval_type import RagasEvalTypeId
from athina.metrics.metric_type import MetricType
from ragas.metrics import answer_correctness

"""
RAGAS Answer Correctness Docs: https://docs.ragas.io/en/latest/concepts/metrics/answer_correctness.html
RAGAS Answer Correctness Github: https://github.com/explodinggradients/ragas/blob/main/src/ragas/metrics/_answer_correctness.py
"""


class RagasAnswerCorrectness(RagasEvaluator):
    """
    This evaluator involves gauging the accuracy of the generated llm response when compared to the ground truth
    """

    @property
    def name(self):
        return RagasEvalTypeId.RAGAS_ANSWER_CORRECTNESS.value

    @property
    def display_name(self):
        return "Ragas Answer Correctness"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.RAGAS_ANSWER_CORRECTNESS.value]

    @property
    def ragas_metric(self):
        return answer_correctness

    @property
    def ragas_metric_name(self):
        return "answer_correctness"

    @property
    def default_model(self):
        return Model.GPT35_TURBO.value

    @property
    def required_args(self):
        return ["query", "response", "expected_response"]

    @property
    def examples(self):
        return None

    @property
    def grade_reason(self) -> str:
        return "Answer correctness encompasses two critical aspects: semantic similarity between the generated answer and the ground truth, as well as factual similarity. These aspects are combined using a weighted scheme to formulate the answer correctness score"

    def is_failure(self, score) -> Optional[bool]:
        return (
            bool(score < self._failure_threshold)
            if self._failure_threshold is not None
            else None
        )

    def generate_data_to_evaluate(
        self, query, response, expected_response, **kwargs
    ) -> dict:
        """
        Generates data for evaluation.

        :param query: user query
        :param response: llm response
        :param expected_response: expected output
        :return: A dictionary with formatted data for evaluation
        """
        data = {
            "question": [query],
            "answer": [response],
            "ground_truths": [[expected_response]],
        }
        return data



from datetime import datetime, timezone
import random
import string


def generate_unique_dataset_name(prefix="Dataset-", separator="-"):
    """Generates a unique name using the current timestamp with separators for readability.

    Args:
        prefix (str): Optional. A prefix for the generated name.
        separator (str): The separator to use between date and time components.

    Returns:
        str: A unique name based on the current timestamp
    """
    # Get the current datetime with desired precision
    current_time = datetime.now()
    # Format the datetime into a string with separators
    time_str = current_time.strftime(
        f"%Y{separator}%m{separator}%d{separator}%H{separator}%M{separator}%S"
    )
    # Combine the prefix and the formatted time string to create a unique name
    return prefix + time_str


def generate_eval_display_name(eval_display_name: str) -> str:
    # Get current UTC timestamp in human-readable format
    timestamp = datetime.now(timezone.utc).strftime("%B%d_%Y_%H%M%S")

    # Generate a random suffix
    random_suffix = "".join(random.choices(string.ascii_uppercase + string.digits, k=3))

    # Combine to form the display name
    eval_display_name = f"{eval_display_name}_{timestamp}_{random_suffix}"

    return eval_display_name



from .metric import Metric


class RagasMetric(Metric):
    """
    Float ragas metric
    """

    @staticmethod
    def compute(value: float):
        """
        Computes the result.

        Returns:
            float: Returns the metric
        """
        return value



data = [
    {
        "document": """
Alice (Veterinarian): Hi Bob, I understand you're looking to get a new dog. It's great that you're considering different breeds like Golden Retrievers, Labradors, and French Bulldogs. Let's discuss what you're looking for in a dog to help you make the best choice.

Bob: Thanks, Alice. I'm really looking for a breed that's hypoallergenic and doesn't shed much. Also, it's important that the dog is friendly and non-aggressive towards other people.

Alice: Those are important considerations. Let's start with Golden Retrievers and Labradors. Both are known for their friendly and outgoing nature, which means they generally get along well with people and other pets. However, they are not hypoallergenic and do tend to shed quite a bit.

Bob: I see, that's a bit of a concern for me. What about French Bulldogs?

Alice: French Bulldogs are a bit different. They're smaller and have a playful, affectionate personality. They tend to shed less than Golden Retrievers and Labradors, but they're not entirely hypoallergenic either. One thing to note is that they can be quite stubborn, which requires consistent training.

Bob: That's helpful to know. I really need a hypoallergenic breed due to allergies. Are there any breeds you would recommend that fit these criteria?

Alice: Absolutely, Bob. For hypoallergenic and low shedding breeds, you might want to consider Poodles, Bichon Frises, or Portuguese Water Dogs. These breeds are known for their friendly demeanor and are less likely to trigger allergies. They also require regular grooming to maintain their coat and minimize shedding.

Bob: That sounds more like what I'm looking for. I hadn't thought about those breeds. I'll definitely look into them. Thanks for your advice, Alice!

Alice: You're welcome, Bob! Feel free to reach out if you have more questions or need help once you decide on a breed. It's important to choose a dog that fits well with your lifestyle and needs.
        """,
        "response": """
In this conversation, Alice, a veterinarian, and Bob discuss Bob's desire to get a new dog. Bob seeks a hypoallergenic breed that sheds minimally and is friendly. Alice notes that while Golden Retrievers and Labradors are friendly, they aren't hypoallergenic and shed a lot. French Bulldogs are less shedding but also not completely hypoallergenic and can be stubborn. Alice then suggests Poodles, Bichon Frises, or Portuguese Water Dogs as breeds fitting Bob's criteria: hypoallergenic, low shedding, and friendly. Bob appreciates the advice and considers these options. Alice offers further assistance as needed.
        """,
        "questions": [
            "Is Bob interested in getting a new dog?",
            "Does Bob prefer a hypoallergenic dog breed?",
            "Is Bob concerned about dog shedding?",
            "Does Alice suggest that Golden Retrievers are hypoallergenic?",
            "Are Labradors known for minimal shedding?",
            "Does Alice mention that French Bulldogs are completely hypoallergenic?",
            "Are Poodles suggested by Alice as a suitable breed for Bob?",
            "Does Alice recommend Bichon Frises to Bob?",
            "Is a Portuguese Water Dog one of the breeds Alice suggests?",
            "Does Bob decide to get a dog immediately after the conversation?",
        ],
    }
]



from typing import List, Optional
from athina.interfaces.athina import AthinaFilters
from athina.interfaces.data import DataPoint
from athina.services.athina_api_service import AthinaApiService
from .base_loader import BaseLoader
from dataclasses import asdict
import json


class JsonLoader(BaseLoader):
    """
    This class is a data loader for json evals

    Attributes:
        col_actual_json (dict or str): The column name corresponding to the actual JSON.
        col_expected_json (dict or str): The column name corresponding to the expected JSON.
        raw_dataset (dict): The raw dataset as loaded from the source.
        processed_dataset (list): The processed dataset with responses.
    """

    def __init__(
        self,
        col_actual_json: str = "actual_json",
        col_expected_json: str = "expected_json",
    ):
        """
        Initializes the loader with specified or default column names.
        """
        self.col_actual_json = col_actual_json
        self.col_expected_json = col_expected_json
        self._raw_dataset = {}
        self._processed_dataset: List[DataPoint] = []

    def process(self) -> None:
        """
        Transforms the raw data into a structured format. Processes each entry from the raw dataset, and extracts attributes.

        Raises:
            KeyError: If mandatory columns (response) are missing in the raw dataset.
        """
        for raw_instance in self._raw_dataset:
            # Check for mandatory columns in raw_instance
            if self.col_actual_json not in raw_instance:
                raise KeyError(f"'{self.col_actual_json}' not found in provided data.")
            # Create a processed instance with mandatory fields
            processed_instance = {
                # if self.col_actual_json is string then do a json load
                "actual_json": (
                    json.loads(raw_instance[self.col_actual_json])
                    if isinstance(raw_instance[self.col_actual_json], str)
                    else raw_instance[self.col_actual_json]
                )
            }
            if self.col_expected_json in raw_instance:
                processed_instance["expected_json"] = (
                    json.loads(raw_instance[self.col_expected_json])
                    if isinstance(raw_instance[self.col_expected_json], str)
                    else raw_instance[self.col_expected_json]
                )
            # removing keys with None values
            processed_instance = {
                k: v for k, v in processed_instance.items() if v is not None
            }
            # Store the results
            self._processed_dataset.append(processed_instance)

    def load_athina_inferences(
        self,
        filters: Optional[AthinaFilters] = None,
        limit: Optional[int] = None,
    ):
        """
        Load data from Athina API.
        """
        raise NotImplementedError(
            "This loader does not support loading data from Athina API."
        )



# Step to make an external api call
import json
import time
from typing import Union, Dict, Any, Iterable, Optional
import requests
from athina.steps import Step
from jinja2 import Environment
from athina.helpers.jinja_helper import PreserveUndefined
import urllib.parse


def prepare_input_data(data):
    return {
        key: json.dumps(value) if isinstance(value, (list, dict)) else value
        for key, value in data.items()
    }


class ApiCall(Step):
    """
    Step that makes an external API call.

    Attributes:
        url: The URL of the API endpoint to call.
        method: The HTTP method to use (e.g., 'GET', 'POST', 'PUT', 'DELETE').
        headers: Optional headers to include in the API request.
        params: Optional params to include in the API request.
        body: Optional request body to include in the API request.
    """

    url: str
    method: str
    headers: Optional[Dict[str, str]] = None
    params: Optional[Dict[str, str]] = None
    body: Optional[str] = None
    env: Environment = None
    name: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True

    def execute(self, input_data: Any) -> Union[Dict[str, Any], None]:
        """Make an API call and return the response."""

        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict):
            raise TypeError("Input data must be a dictionary.")

        # Create a custom Jinja2 environment with double curly brace delimiters and PreserveUndefined
        self.env = Environment(
            variable_start_string="{{",
            variable_end_string="}}",
            undefined=PreserveUndefined,
        )
        prepared_body = None
        # Add a filter to the Jinja2 environment to convert the input data to JSON
        if self.body is not None:
            body_template = self.env.from_string(self.body)
            prepared_input_data = prepare_input_data(input_data)
            prepared_body = body_template.render(**prepared_input_data)

        prepared_headers = self.headers.copy() if self.headers is not None else None
        prepared_params = self.params.copy() if self.params is not None else None

        if prepared_headers is not None:
            for key, value in prepared_headers.items():
                prepared_headers[key] = self.env.from_string(value).render(
                    **prepared_input_data
                )

        if prepared_params is not None:
            for key, value in prepared_params.items():
                prepared_params[key] = self.env.from_string(value).render(
                    **prepared_input_data
                )

        retries = 2  # number of retries
        timeout = 30  # seconds
        for attempt in range(retries):
            try:
                response = requests.request(
                    method=self.method,
                    url=self.url,
                    headers=prepared_headers,
                    params=prepared_params,
                    json=(
                        json.loads(prepared_body, strict=False)
                        if prepared_body
                        else None
                    ),
                    timeout=timeout,
                )
                if response.status_code >= 400:
                    # If the status code is an error, return the error message
                    return {
                        "status": "error",
                        "data": f"Failed to make the API call.\nStatus code: {response.status_code}\nError:\n{response.text}",
                    }
                try:
                    json_response = response.json()
                    # If the response is JSON, return the JSON data
                    return {
                        "status": "success",
                        "data": json_response,
                    }
                except json.JSONDecodeError:
                    # If the response is not JSON, return the text
                    return {
                        "status": "success",
                        "data": response.text,
                    }
            except requests.Timeout:
                if attempt < retries - 1:
                    time.sleep(2)
                    continue
                # If the request times out after multiple attempts, return an error message
                return {
                    "status": "error",
                    "data": "Failed to make the API call.\nRequest timed out after multiple attempts.",
                }
            except Exception as e:
                # If an exception occurs, return the error message
                return {
                    "status": "error",
                    "data": f"Failed to make the API call.\nError: {e.__class__.__name__}\nDetails:\n{str(e)}",
                }



from .metric import Metric


class AgreementScore(Metric):
    """
    Calculates agreement score between two sets of answers.

    AgreementScore computes the proportion of questions that received
    consistent answers between a source (e.g., document) and a summary.
    """

    @staticmethod
    def _compute_metric(answers_src, answers_sum, questions):
        """
        Computes the number of matches between the answers from source and summary.

        Args:
            answers_src (dict): Answers derived from the source.
            answers_sum (dict): Answers derived from the summary.

        Returns:
            int: Number of questions with consistent answers.
        """
        answers_src_ls = list(answers_src.values())
        answers_sum_ls = list(answers_sum.values())
        n_matches = 0
        agreed_questions = []
        for idx, (ans_src, ans_sum) in enumerate(zip(answers_src_ls, answers_sum_ls)):
            if ans_src.strip().lower() == ans_sum.strip().lower():
                n_matches += 1
                agreed_question = questions[idx]
                agreed_questions.append(f"{agreed_question}")
        return n_matches, agreed_questions

    @staticmethod
    def compute(answers_src, answers_sum, questions, n_questions):
        """
        Computes the agreement score.

        Args:
            answers_src (dict): Answers derived from the source.
            answers_sum (dict): Answers derived from the summary.
            n_questions (int): Total number of questions.

        Returns:
            float: Agreement score.
        """
        n_matches, agreed_questions = AgreementScore._compute_metric(
            answers_src, answers_sum, questions
        )
        explanation = agreed_questions
        agreement_score = n_matches / n_questions
        return agreement_score, explanation
